@@ -7,14 +7,12 @@ Matlab version:
     course: Multivariable Control Systems
 """
 import numpy as np
-
-from md_reg import * # needed for vif()
+from aislab.md_reg import * # needed for vif()
+from aislab.dp_feng.normlz import *
+from aislab.gnrl.bf import *
 
 
-
-
-   
-# def assoc(model = None,st = None,y = None,ym = None): 
+# def assoc(model = None,st = None,y = None,ym = None):
 #     #--------------------------
 # # Author: Alexander Efremov
 # # Date:   20.06.2011
@@ -98,7 +96,7 @@ def auc(g, b):
     pg = g/G*100
     pb = b/B*100
     cpg = np.cumsum(pg, axis=0)
-    tmp = pb*(cpg + np.vstack( (np.zeros((1, n)), cpg[:-1, :]) ))
+    tmp = pb*(cpg + np.vstack((np.zeros((1, n)), cpg[:-1, :])))
     gini = abs(100 - 0.005*tmp.sum(axis=0))
     return gini
 ###################################################################################
@@ -154,7 +152,6 @@ def betai(a, b, x):
     a = c_(a)
     b = c_(b)
     x = c_(x)
-    
     bi = nans(x.shape)
     if ((x < 0) | (x > (1 + 1e-14))).any():
         print('ERROR: SpecialFunctions, betai():', 'x not in [0, 1].')
@@ -181,14 +178,24 @@ def chi2(g, b):
     ebi = (g + b)*B/(G + B)
     egi[egi < 1e-12] = 1e-12
     ebi[ebi < 1e-12] = 1e-12
-    chi = ((g - egi)**2/egi + (b - ebi)**2/ebi).sum(axis = 0).T
+    chi = ((g - egi)**2/egi + (b - ebi)**2/ebi).sum(axis=0).T
     return chi
 ###################################################################################
+def confm(x1, x2, w=None, ux1=None, ux2=None):
+    # confusion matrix
+    if ux1 is None: ux1 = np.unique(x1)
+    if ux2 is None: ux2 = np.unique(x2)
+    cm = nans((len(ux1), len(ux2)))
+    for i in range(len(ux1)):
+        for j in range(len(ux2)):
+            cm[i, j] = np.sum(w[(x1 == ux1[i]) & (x2 == ux2[j])])
+    return cm
+###################################################################################
 def corf(x, y=None):
     if y is None: y = x
     N = x.shape[0]
     X = hnkl(np.vstack((x, x)), N)
-    R = X@y/(N - 1)
+    R = X @ y / (N - 1)
     return R
 ###################################################################################
 def covf(x, y=None):
@@ -202,14 +209,16 @@ def dind(g, b):
     B = sum(b)
     gb = g + b
     gb[gb < 1e-12] = 1e-12
-    dind = 100 - sum((g*b/gb*(G + B)/(G*B)))*100
+    dind = 100 - sum((g * b / gb * (G + B) / (G * B))) * 100
     return dind
 ##############################################################################
 def frat(y, y2, nw):
-    if y.ndim == 1 and len(y) < 1: f = np.nan; return(f)
-    if y.ndim == 1: m, n = len(y), 1;
-    else:           m, n = y.shape
-    eps = 1e-6# np.finfo(float).eps
+    if y.ndim == 1 and len(y) < 1: f = np.nan; return f
+    if y.ndim == 1:
+        m, n = len(y), 1
+    else:
+        m, n = y.shape
+    eps = 1e-6  # np.finfo(float).eps
     nw[nw < eps] = eps
     ssy = sum(y2)
     rc = sum(y.astype('float')**2/nw)
@@ -222,19 +231,21 @@ def frat(y, y2, nw):
     return f
 ###################################################################################
 def gamln(a):
-    cof = c_(np.array([76.18009172947146, -86.50532032941677, 24.01409824083091, -1.231739572450155, 0.1208650973866179e-2, 0.5395239384953e-5]))
-    y = a + c_(np.arange(1,7))
+    cof = c_(np.array(
+        [76.18009172947146, -86.50532032941677, 24.01409824083091, -1.231739572450155, 0.1208650973866179e-2,
+         0.5395239384953e-5]))
+    y = a + c_(np.arange(1, 7))
     tmp = a + 5.5
-    tmp = tmp - (a + 0.5)*np.log(tmp)
-    ser = 1.000000000190015 + sum(cof/y)
-    return -tmp + np.log(2.5066282746310005*ser/a)
+    tmp = tmp - (a + 0.5) * np.log(tmp)
+    ser = 1.000000000190015 + sum(cof / y)
+    return -tmp + np.log(2.5066282746310005 * ser / a)
 ###################################################################################
 def gamq(a, x):
     # a must be real and non-negative
     if type(a) != np.array: a = np.array([[a]])
     if type(x) != np.array: x = np.array([[x]])
-	
-	# if x.min() < 0 or a.min() <= 0:
+
+    # if x.min() < 0 or a.min() <= 0:
     #     print('ERROR: SpecialFunctions, gamq():', 'Invalid arguments in routine gamq')
     #     return -1
     if a.shape == (1, 1): a = np.tile(a, x.shape)
@@ -245,8 +256,10 @@ def gamq(a, x):
     gq[(a > 0) & (x == 0)] = 1
     for i in range(0, len(gq.flatten())):
         if x[i] <= 0 or a[i] <= 1e-300: continue
-        if x[i] < a[i] + 1: gq[i] = 1 - gser(a[i], x[i])
-        else:               gq[i] = gcf(a[i], x[i])
+        if x[i] < a[i] + 1:
+            gq[i] = 1 - gser(a[i], x[i])
+        else:
+            gq[i] = gcf(a[i], x[i])
     return gq
 ###################################################################################
 def gbi(g, b):
@@ -255,15 +268,18 @@ def gbi(g, b):
     G = np.tile(g.sum(axis=0), (m, 1))
     B = np.tile(b.sum(axis=0), (m, 1))
     b[b < 1e-12] = 1e-12
-    gbo = g/b; GBO = G/B
+    gbo = g / b;
+    GBO = G / B
     gbo = gbo.flatten()
     GBO = GBO.flatten()
     gbi = np.zeros(len(gbo))
-    ind = gbo >= GBO;  gbi[ind] = gbo[ind]/GBO[ind]*100
+    ind = gbo >= GBO;
+    gbi[ind] = gbo[ind] / GBO[ind] * 100
     gbo[gbo < 1e-12] = 1e-12
-    ind = gbo < GBO;   gbi[ind] = -GBO[ind]/gbo[ind]*100
+    ind = gbo < GBO;
+    gbi[ind] = -GBO[ind] / gbo[ind] * 100
     gbi = gbi.reshape(m, n)
-    gbi1 = gbi + 200*(gbi < 0).astype(int)
+    gbi1 = gbi + 200 * (gbi < 0).astype(int)
     gbi1 = gbi1.reshape(m, n)
     return gbi
 ###################################################################################
@@ -275,21 +291,26 @@ def gcf(a, x, delta=1e-7, maxIter=int(1e3)):
         print('ERROR: SpecialFunctions, gcf():', 'Wrong gamma-function parameter')
         return -1
     a1 = x
-    N = range(1, maxIter+1)
+    N = range(1, maxIter + 1)
     for n in N:
-        an = n;                  ana = an - a
-        a0 = (a1 + a0*ana)*fac;  b0 = (b1 + b0*ana)*fac
-        anf = an*fac
-        a1 = x*a0 + anf*a1;      b1 = x*b0 + anf*b1
+        an = n;
+        ana = an - a
+        a0 = (a1 + a0 * ana) * fac;
+        b0 = (b1 + b0 * ana) * fac
+        anf = an * fac
+        a1 = x * a0 + anf * a1;
+        b1 = x * b0 + anf * b1
         if a1 != 0:
-            fac = 1/a1
-            g = b1*fac
-            if abs((g - gold)/g) < delta:
-                temp = -x + a*np.log(x) - gln
+            fac = 1 / a1
+            g = b1 * fac
+            if abs((g - gold) / g) < delta:
+                temp = -x + a * np.log(x) - gln
                 if abs(temp) > 700:
-                    if temp < -700: temp = -700
-                    else:           temp =  700
-                gcf = np.exp(temp)*g
+                    if temp < -700:
+                        temp = -700
+                    else:
+                        temp = 700
+                gcf = np.exp(temp) * g
                 break
             gold = g
     return gcf
@@ -299,11 +320,11 @@ def gini(g, b):
     [m, n] = g.shape
     G = g.sum(axis=0)
     B = b.sum(axis=0)
-    pg = g/G*100
-    pb = b/B*100
+    pg = g / G * 100
+    pb = b / B * 100
     cpg = np.cumsum(pg, axis=0)
-    tmp = pb*(cpg + np.vstack( (np.zeros((1, n)), cpg[:-1, :]) ))
-    gini = abs(100 - 0.01*tmp.sum(axis=0))
+    tmp = pb * (cpg + np.vstack((np.zeros((1, n)), cpg[:-1, :])))
+    gini = abs(100 - 0.01 * tmp.sum(axis=0))
     return gini
 ###################################################################################
 def gser(a, x):
@@ -315,11 +336,15 @@ def gser(a, x):
         return -1
     gln = gamln(a)
     gamser = 0
-    ap = a;  inva = 1/a;  s = inva
+    ap = a;
+    inva = 1 / a;
+    s = inva
     for n in range(0, maxIter):
-        ap = ap + 1;  inva = inva*x/ap;  s = s + inva
-        if abs(inva) < abs(s)*Eps:
-            gamser = s*np.exp(-x + a*np.log(x) - gln)
+        ap = ap + 1;
+        inva = inva * x / ap;
+        s = s + inva
+        if abs(inva) < abs(s) * Eps:
+            gamser = s * np.exp(-x + a * np.log(x) - gln)
             break
     return gamser
 ###################################################################################
@@ -328,37 +353,42 @@ def ks(g, b):
     [m, n] = g.shape
     G = g.sum(axis=0)
     B = b.sum(axis=0)
-    pg = g/G*100
-    pb = b/B*100
+    pg = g / G * 100
+    pb = b / B * 100
     cpg = np.cumsum(pg, axis=0)
     cpb = np.cumsum(pb, axis=0)
     ks = np.max(abs(cpg - cpb), axis=0)
     return ks
+
+
 ###################################################################################
 def pvalC(v, C2):
-    return gamq(0.5*v, 0.5*C2)
+    return gamq(0.5 * v, 0.5 * C2)
 ###################################################################################
 def pvalF(F, v1, v2, tp='ot'):
-    #Computes the p-value for Fisher
-    #Inputs:
+    # Computes the p-value for Fisher
+    # Inputs:
     #   F - F value
     #   v1 - regression degeres of freedom
     #   v2 - error degeres of freedom
     #   tp - type of test /default is 'ot'/
     #      'tt' - two tail
     #      'ot' - one tail
-    #Output
+    # Output
     #   p - p-value of F value
     F = c_(F)
-    if v2 <= 0 or v2 <= 0: print('ERROR: SpecialFunctions, pvalF():', 'Degrees of freedom must be positive...'); return []
+    if v2 <= 0 or v2 <= 0: print('ERROR: SpecialFunctions, pvalF():',
+                                 'Degrees of freedom must be positive...'); return []
     if (F < 0).any():      print('ERROR: SpecialFunctions, pvalF():', 'x must be non negative...'); return []
-    if tp == 'ot': p = betai(v2/2, v1/2, v2/(v2 + F*v1))
-    else:          p = 2*betai(v2/2, v1/2, v2/(v2 + F*v1))
-    p[p < 0] = 0 # p may be out of range due to round-off errors
-    p[p > 1] = 1 # 
+    if tp == 'ot':
+        p = betai(v2 / 2, v1 / 2, v2 / (v2 + F * v1))
+    else:
+        p = 2 * betai(v2 / 2, v1 / 2, v2 / (v2 + F * v1))
+    p[p < 0] = 0  # p may be out of range due to round-off errors
+    p[p > 1] = 1  #
     return p
 ###################################################################################
-def vaf(Y, Ym, w=np.empty([0,0]), p=0):
+def vaf(Y, Ym, w=np.empty([0, 0]), p=0):
     Nw, r = Y.shape
     if w.size > 0:
         if w.size < Y.size: w = np.tile(w, (1, r))
@@ -370,9 +400,9 @@ def vaf(Y, Ym, w=np.empty([0,0]), p=0):
         mY = np.mean(Y, axis=0)
         SSE = sum((Y - Ym)**2)
         SST = sum((Y - mY)**2)
-    VAF = 1 - (SSE/SST*(Nw - 1)/(Nw - p - 1))
+    VAF = 1 - (SSE/ SST*(Nw - 1)/(Nw - p - 1))
     VAF[VAF < 0] = 0
-    return VAF*100
+    return VAF * 100
 ###################################################################################
 def vif(x):
     m = x.shape[1]
@@ -382,67 +412,30 @@ def vif(x):
         xi = np.delete(x, i, axis=1)
         mdl = lspm(xi, yi)
         yim = lspm_apl(xi, yi, mdl=mdl)
-        R2 = vaf(yi, yim)/100
-        if R2 == 1: vifi[i] = np.inf
-        else:       vifi[i] = 1/(1 - R2)
+        R2 = vaf(yi, yim) / 100
+        if R2 == 1:
+            vifi[i] = np.inf
+        else:
+            vifi[i] = 1 / (1 - R2)
     return vifi
 ###################################################################################
 def vinf(g, b, tau=1e-12):
     G = sum(g)
     B = sum(b)
-    gG = g/G
-    bB = b/B
+    gG = g / G
+    bB = b / B
     gG[gG < tau] = tau
     bB[bB < tau] = tau
-    vinf = sum((gG - bB)*np.log(gG/bB))
+    vinf = sum((gG - bB) * np.log(gG / bB))
     return vinf
 ###################################################################################
 def woe(g, b, tau=1e-12):
     G = sum(g)
     B = sum(b)
-    gG = g/G
-    bB = b/B
+    gG = g / G
+    bB = b / B
     gG[gG < tau] = tau
     bB[bB < tau] = tau
-    woe = np.log(gG/bB)
+    woe = np.log(gG / bB)
     return woe
 ##############################################################################
-##############################################################################
-##############################################################################
-# python equivalent of matlabs' tic & toc functions
-# taken from: https://stackoverflow.com/questions/5849800/what-is-the-python-equivalent-of-matlabs-tic-and-toc-functions
-import time
-def TicTocGenerator():
-    # Generator that returns time differences
-    ti = 0           # initial time
-    tf = time.time() # final time
-    while True:
-        ti = tf
-        tf = time.time()
-        yield tf-ti # returns the time difference
-
-TicToc = TicTocGenerator() # create an instance of the TicTocGen generator
-
-# This will be the main function through which we define both tic() and toc()
-def toc(dsp=1, tempBool=True):
-    # Prints the time difference yielded by generator instance TicToc
-    tempTimeInterval = next(TicToc)
-    if tempBool:
-        if dsp: print( "Elapsed time: %f seconds.\n" %tempTimeInterval )
-        return tempTimeInterval
-
-def tic():
-    # Records a time in TicToc, marks the beginning of a time interval
-    toc(False)
-TicToc2 = TicTocGenerator() # create another instance of the TicTocGen generator
-
-def toc2(dsp=1, tempBool=True):
-    # Prints the time difference yielded by generator instance TicToc2
-    tempTimeInterval = next(TicToc2)
-    if tempBool:
-        if dsp: print("Elapsed time 2: %f seconds.\n" %tempTimeInterval)
-        return tempTimeInterval   
-def tic2():
-    # Records a time in TicToc2, mar
-    toc2(False)
-##############################################################################
